@@ -36,6 +36,7 @@
 #include <QDebug>
 
 #include <kauthorized.h>
+#include <klocalizedstring.h>
 
 #include <assert.h>
 
@@ -48,6 +49,9 @@ public:
           m_parent(0L),
           m_builder(0L)
     {
+        m_textTagNames.append(QStringLiteral("text"));
+        m_textTagNames.append(QStringLiteral("Text"));
+        m_textTagNames.append(QStringLiteral("title"));
     }
     ~KXMLGUIClientPrivate()
     {
@@ -73,6 +77,7 @@ public:
     KXMLGUIBuilder *m_builder;
     QString m_xmlFile;
     QString m_localXMLFile;
+    QStringList m_textTagNames;
 
     // Actions to enable/disable on a state change
     QMap<QString, KXMLGUIClient::StateChange> m_actionsStateMap;
@@ -194,15 +199,20 @@ void KXMLGUIClient::setComponentName(const QString &componentName, const QString
     }
 }
 
+QString KXMLGUIClient::standardsXmlFileLocation()
+{
+    QString file = QStandardPaths::locate(QStandardPaths::GenericConfigLocation, QStringLiteral("ui/ui_standards.rc"));
+    if (file.isEmpty()) {
+        // fallback to resource, to allow to use the rc file compiled into this framework, must exist!
+        file = QStringLiteral(":/kxmlgui5/ui_standards.rc");
+        Q_ASSERT(QFile::exists(file));
+    }
+    return file;
+}
+
 void KXMLGUIClient::loadStandardsXmlFile()
 {
-    const QString file = QStandardPaths::locate(QStandardPaths::GenericConfigLocation, QStringLiteral("ui/ui_standards.rc"));
-    if (file.isEmpty()) {
-        qWarning() << "ui/ui_standards.rc not found in" << QStandardPaths::standardLocations(QStandardPaths::GenericConfigLocation);
-    } else {
-        const QString doc = KXMLGUIFactory::readConfigFile(file);
-        setXML(doc);
-    }
+    setXML(KXMLGUIFactory::readConfigFile(standardsXmlFileLocation()));
 }
 
 void KXMLGUIClient::setXMLFile(const QString &_file, bool merge, bool setXMLDoc)
@@ -222,10 +232,21 @@ void KXMLGUIClient::setXMLFile(const QString &_file, bool merge, bool setXMLDoc)
         allFiles.append(file);
     } else {
         const QString filter = componentName() + QLatin1Char('/') + _file;
-        allFiles = QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("kxmlgui5/") + filter); // KF >= 5.1
+
+        // files on filesystem
+        allFiles << QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, QStringLiteral("kxmlgui5/") + filter); // KF >= 5.1
+
+        // KF >= 5.4 (resource file)
+        const QString qrcFile(QStringLiteral(":/kxmlgui5/") + filter);
+        if (QFile::exists(qrcFile)) {
+            allFiles << qrcFile;
+        }
+
+        // then compat locations
         const QStringList compatFiles =
                    QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, filter) + // kdelibs4, KF 5.0
                    QStandardPaths::locateAll(QStandardPaths::GenericDataLocation, _file); // kdelibs4, KF 5.0, caller passes component name
+
         if (allFiles.isEmpty() && !compatFiles.isEmpty()) {
             qWarning() << "KXMLGUI file found at deprecated location" << compatFiles << "-- please use ${KXMLGUI_INSTALL_DIR} to install this file instead.";
         }
@@ -269,15 +290,46 @@ void KXMLGUIClient::replaceXMLFile(const QString &xmlfile, const QString &localx
     setXMLFile(xmlfile, merge);
 }
 
+// The top document element may have translation domain attribute set,
+// or the translation domain may be implicitly the application domain.
+// This domain must be used to fetch translations for all text elements
+// in the document that do not have their own domain attribute.
+// In order to preserve this semantics through document mergings,
+// the top or application domain must be propagated to all text elements
+// lacking their own domain attribute.
+static void propagateTranslationDomain(QDomDocument &doc, const QStringList tagNames)
+{
+    const QLatin1String attrDomain("translationDomain");
+    QDomElement base = doc.documentElement();
+    QString domain = base.attribute(attrDomain);
+    if (domain.isEmpty()) {
+        domain = QString::fromUtf8(KLocalizedString::applicationDomain());
+        if (domain.isEmpty()) {
+            return;
+        }
+    }
+    foreach (const QString &tagName, tagNames) {
+        QDomNodeList textNodes = base.elementsByTagName(tagName);
+        for (int i = 0; i < textNodes.length(); ++i) {
+            QDomElement e = textNodes.item(i).toElement();
+            QString localDomain = e.attribute(attrDomain);
+            if (localDomain.isEmpty()) {
+                 e.setAttribute(attrDomain, domain);
+            }
+        }
+    }
+}
+
 void KXMLGUIClient::setXML(const QString &document, bool merge)
 {
     QDomDocument doc;
     QString errorMsg;
-    int errorLine, errorColumn;
+    int errorLine = 0, errorColumn = 0;
     // QDomDocument raises a parse error on empty document, but we accept no app-specific document,
     // in which case you only get ui_standards.rc layout.
     bool result = document.isEmpty() || doc.setContent(document, &errorMsg, &errorLine, &errorColumn);
     if (result) {
+        propagateTranslationDomain(doc, d->m_textTagNames);
         setDOMDocument(doc, merge);
     } else {
 #ifdef NDEBUG
@@ -304,7 +356,7 @@ void KXMLGUIClient::setDOMDocument(const QDomDocument &document, bool merge)
         // strange to it
         base = d->m_doc.documentElement();
 
-        //qDebug(260) << "Result of xmlgui merging:" << d->m_doc.toString();
+        //qCDebug(DEBUG_KXMLGUI) << "Result of xmlgui merging:" << d->m_doc.toString();
 
         // we want some sort of failsafe.. just in case
         if (base.isNull()) {
@@ -318,7 +370,11 @@ void KXMLGUIClient::setDOMDocument(const QDomDocument &document, bool merge)
 }
 
 // if (equals(a,b)) is more readable than if (a.compare(b, Qt::CaseInsensitive)==0)
-static bool equalstr(const QString &a, const QString &b)
+static inline bool equalstr(const QString &a, const QString &b)
+{
+    return a.compare(b, Qt::CaseInsensitive) == 0;
+}
+static inline bool equalstr(const QString &a, const QLatin1String &b)
 {
     return a.compare(b, Qt::CaseInsensitive) == 0;
 }
@@ -328,13 +384,13 @@ bool KXMLGUIClientPrivate::mergeXML(QDomElement &base, QDomElement &additive, KA
     const QLatin1String tagAction("Action");
     const QLatin1String tagMerge("Merge");
     const QLatin1String tagSeparator("Separator");
-    const QLatin1String attrName("name");
-    const QLatin1String attrAppend("append");
-    const QLatin1String attrWeakSeparator("weakSeparator");
     const QLatin1String tagMergeLocal("MergeLocal");
     const QLatin1String tagText("text");
-    const QLatin1String attrAlreadyVisited("alreadyVisited");
-    const QLatin1String attrNoMerge("noMerge");
+    const QLatin1String attrAppend("append");
+    const QString       attrName(QStringLiteral("name"));
+    const QString       attrWeakSeparator(QStringLiteral("weakSeparator"));
+    const QString       attrAlreadyVisited(QStringLiteral("alreadyVisited"));
+    const QString       attrNoMerge(QStringLiteral("noMerge"));
     const QLatin1String attrOne("1");
 
     // there is a possibility that we don't want to merge in the
@@ -450,7 +506,7 @@ bool KXMLGUIClientPrivate::mergeXML(QDomElement &base, QDomElement &additive, KA
             else {
                 QDomElement matchingElement = findMatchingElement(e, additive);
                 if (!matchingElement.isNull()) {
-                    matchingElement.setAttribute(attrAlreadyVisited, (uint)1);
+                    matchingElement.setAttribute(attrAlreadyVisited, uint(1));
 
                     if (mergeXML(e, matchingElement, actionCollection)) {
                         base.removeChild(e);
@@ -518,14 +574,14 @@ bool KXMLGUIClientPrivate::isEmptyContainer(const QDomElement &base, KActionColl
 
         const QString tag = e.tagName();
 
-        if (equalstr(tag, QStringLiteral("Action"))) {
+        if (equalstr(tag, QLatin1String("Action"))) {
             // if base contains an implemented action, then we must not get
             // deleted (note that the actionCollection contains both,
             // "global" and "local" actions)
             if (actionCollection->action(e.attribute(QStringLiteral("name")))) {
                 return false;
             }
-        } else if (equalstr(tag, QStringLiteral("Separator"))) {
+        } else if (equalstr(tag, QLatin1String("Separator"))) {
             // if we have a separator which has *not* the weak attribute
             // set, then it must be owned by the "local" tree in which case
             // we must not get deleted either
@@ -535,12 +591,12 @@ bool KXMLGUIClientPrivate::isEmptyContainer(const QDomElement &base, KActionColl
             }
         }
 
-        else if (equalstr(tag, QStringLiteral("merge"))) {
+        else if (equalstr(tag, QLatin1String("merge"))) {
             continue;
         }
 
         // a text tag is NOT enough to spare this container
-        else if (equalstr(tag, QStringLiteral("text"))) {
+        else if (equalstr(tag, QLatin1String("text"))) {
             continue;
         }
 
@@ -568,14 +624,14 @@ QDomElement KXMLGUIClientPrivate::findMatchingElement(const QDomElement &base, c
 
         const QString tag = e.tagName();
         // skip all action and merge tags as we will never use them
-        if (equalstr(tag, QStringLiteral("Action"))
-                || equalstr(tag, QStringLiteral("MergeLocal"))) {
+        if (equalstr(tag, QLatin1String("Action"))
+                || equalstr(tag, QLatin1String("MergeLocal"))) {
             continue;
         }
 
         // now see if our tags are equivalent
         if (equalstr(tag, base.tagName()) &&
-                e.attribute(QStringLiteral("name")) == base.attribute(QLatin1String("name"))) {
+                e.attribute(QStringLiteral("name")) == base.attribute(QStringLiteral("name"))) {
             return e;
         }
     }
@@ -679,7 +735,7 @@ void KXMLGUIClient::addStateActionEnabled(const QString &state,
     StateChange stateChange = getActionsToChangeForState(state);
 
     stateChange.actionsToEnable.append(action);
-    //qDebug(260) << "KXMLGUIClient::addStateActionEnabled( " << state << ", " << action << ")";
+    //qCDebug(DEBUG_KXMLGUI) << "KXMLGUIClient::addStateActionEnabled( " << state << ", " << action << ")";
 
     d->m_actionsStateMap.insert(state, stateChange);
 }
@@ -690,7 +746,7 @@ void KXMLGUIClient::addStateActionDisabled(const QString &state,
     StateChange stateChange = getActionsToChangeForState(state);
 
     stateChange.actionsToDisable.append(action);
-    //qDebug(260) << "KXMLGUIClient::addStateActionDisabled( " << state << ", " << action << ")";
+    //qCDebug(DEBUG_KXMLGUI) << "KXMLGUIClient::addStateActionDisabled( " << state << ", " << action << ")";
 
     d->m_actionsStateMap.insert(state, stateChange);
 }
